﻿/**
 * Metasequoia.as
 *
 * @see http://snippets.libspark.org/
 * @see http://www.libspark.org/wiki/nanmo/Metasequoia
 *
 * History
 * Make Metasequoia.as(original) by rch850
 *  in http://snippets.libspark.org/trac/wiki/rch850/Metasequoia
 * Modify Metasequoia.as for Away3D by katopz@sleepydesign.com
 *  in http://sleepydesign.blogspot.com/2008/05/away3d-metasequoia-mqo.html
 * Modify Metasequoia.as for Away3D v2.2 by nanmo
 *  in http://www.libspark.org/wiki/nanmo/Metasequoia
*/
package away3d.loaders
{
	import away3d.containers.ObjectContainer3D;
    import away3d.core.*;
    import away3d.core.math.*;
    import away3d.core.base.*
    import away3d.core.base.*;
	import away3d.loaders.data.FaceData;
	import away3d.loaders.data.MeshData;
	import away3d.loaders.data.MeshMaterialData;
	import away3d.loaders.utils.MaterialLibrary;
	import away3d.loaders.data.MaterialData;
	import away3d.loaders.data.GeometryData;
    import away3d.materials.*;
    import away3d.core.utils.*;
    import away3d.core.stats.*;
	import away3d.loaders.*;	
	import away3d.arcane;	
    import flash.utils.*;

	use namespace arcane;
	
	public class Metasequoia extends AbstractParser 
	{
		/** @private */
        arcane override function prepareData(data:*):void
        {
			//mqo = ByteArray(data);
			mqo = Cast.bytearray(data);

			var lines:Array = mqo.readMultiByte(mqo.length, charset).split("\r\n");
			var l:int = 0;

			// parse materials
			l = parseMaterialChunk(lines, 0);

			// parse objects
			while (l != -1)
			{
				l = parseObjectChunk(lines, l);
			}

			//build materials
			buildMaterials();			
			
			Debug.trace("[Metasequoia] Building Meshes");
			
			//build the meshes			
			for each (_meshData in _meshDataList)
			{
				buildMesh(_meshData, _container as ObjectContainer3D);			
			}
		}
		
		/** An array of bytes from the mqo files. */
		private var mqo:ByteArray;		
		private var _filename:String;		
		private var _meshData:MeshData;
		private var _geometryData:GeometryData;
		private var _moveVector:Number3D = new Number3D();
		private var _materialNames:Array = new Array();
		private var _materialArray:Array;

		/**
		 * Array of mesh data objects used for storing the parsed data structure.
		 */
		private var _meshDataList:Array = [];

		private function getMaterialChunkLine(lines:Array, startLine:int = 0):int 
		{
			for (var i:uint = startLine; i < lines.length; ++i) 
			{
				if (lines[i].indexOf("Material") == 0) 
				{
					return int(i);
				}
			}
			return -1;
		}
		
		private function parseMaterialChunk(lines:Array, startLine:int):int 
		{
			var l:int = getMaterialChunkLine(lines, startLine);
			if (l == -1) 
			{
				return -1;
			}
			
			var line:String = lines[l];
			
			var num:Number = parseInt(line.substr(9));
			if (isNaN(num)) {
				return -1;
			}
			++l;
			_materialNames = new Array();
			
			var endLine:int = l + int(num);
			
			//Get Mqo file directory
			var path:String = _filename.slice(0, _filename.lastIndexOf("/") + 1);
			var materialData:MaterialData;
			while (l < endLine)
			{
				line = lines[l];
				
				var nameBeginIndex:int = line.indexOf("\"");
				var nameEndIndex:int = line.indexOf("\"", nameBeginIndex + 1);
				var name:String = line.substring(nameBeginIndex + 1, nameEndIndex);
				
				materialData = _materialLibrary.addMaterial(name);
				_materialNames.push(materialData);
				
				// check texture
				var tex:String = getParam(line, "tex");				
				if (tex) 
				{					
					materialData.materialType = MaterialData.TEXTURE_MATERIAL;

					tex = tex.substr(1, tex.length - 2);

					if (tex.toLowerCase().search(/\.tga$/) != -1 || tex.toLowerCase().search(/\.bmp$/) != -1)
					{
						//TODO : someone really need TGA, BMP on web?
						//_material = loadTGAMaterial(path + tex);

						trace(" ! Error : TGA, BMP not support.");
						trace(" < Try : .png with same name...");

						tex = tex.replace(/\.tga$/, ".png");
						tex = tex.replace(/\.bmp$/, ".png");
					}

					materialData.textureFileName = tex;					
				} 
				else 
				{
					var colorstr:String = getParam(line, "col");
					if (colorstr != null) 
					{
						var color:Array = colorstr.match(/\d+\.\d+/g);
						var r:int = parseFloat(color[0]) * 255;
						var g:int = parseFloat(color[1]) * 255;
						var b:int = parseFloat(color[2]) * 255;
						var a:Number = parseFloat(color[3]);
						
						if (shading)
						{
							materialData.materialType = MaterialData.SHADING_MATERIAL;
						}
						else
						{
							materialData.materialType = MaterialData.COLOR_MATERIAL;
						}
						
						materialData.diffuseColor = (r << 16) | (g << 8) | b;
						
						// TODO how to set material alpha??
						//materialData.alpha = a;
					} 
					else 
					{
						materialData.materialType = MaterialData.WIREFRAME_MATERIAL;					
					}
				}					

				++l;				
			}
			
			return endLine;
		}
		
		private function getObjectChunkLine(lines:Array, startLine:int = 0):int 
		{
			for (var i:uint = startLine; i < lines.length; ++i) 
			{
				if (lines[i].indexOf("Object") == 0) 
				{
					return int(i);
				}
			}
			return -1;
		}
		
		private function parseObjectChunk(lines:Array, startLine:int):int 
		{
			n = -1;
			
            var vertices:Array = [];
			
			var l:int = getObjectChunkLine(lines, startLine);
			if (l == -1) 
			{
				return -1;
			}
			
			var line:String = lines[l];
			
			var name:String = line.substring(8, line.indexOf("\"", 8));
			++l;
			
			// vertex chunks
			var vertexLine:int = getChunkLine(lines, "vertex", l);
			if (vertexLine == -1) 
			{
				return -1;
			}
			
			var properties:Dictionary = new Dictionary();
			while (l < vertexLine) 
			{
				line = lines[l];
				var props:Array = RegExp(/^\s*([\w]+)\s+(.*)$/).exec(line);
				properties[props[1]] = props[2];
				
				++l;
			}
			
			// start getting the vertices
			line = lines[l];
			l = vertexLine + 1;
			
			var numVertices:int = parseInt(line.substring(line.indexOf("vertex") + 7));
			var vertexEndLine:int = l + numVertices;
			var firstVertexIndex:int = vertices.length;			
			while (l < vertexEndLine)
			{
				line = lines[l];
				var coords:Array = line.match(/(-?\d+\.\d+)/g);
				var x:Number = parseFloat(coords[0]) * scaling;
				var y:Number = parseFloat(coords[1]) * scaling;
				var z:Number = -parseFloat(coords[2]) * scaling;
				vertices.push(new Vertex(x, y, z));
				++l;
			}
			
			// start getting  the faces
			l = getChunkLine(lines,  "face", l);
			if (l == -1) 
			{
				return -1;
			}
			line = lines[l++];			
			
			var numFaces:int = parseInt(line.substring(line.indexOf("face") + 5));
			var faceEndLine:int = l + numFaces;
			
			_meshData = new MeshData();
			_meshData.name = name;
			
			// DEBUG
			//trace("[Metaseq] Adding mesh object:" + name);

			_geometryData = _meshData.geometry = _geometryLibrary.addGeometry(_meshData.name);
			
			// parse faces
			while (l < faceEndLine)
			{
				if (properties["visible"] == "15")
				{
					//_meshData.material = parseFace(lines[l], vertices, firstVertexIndex, properties);					
					parseFace(lines[l], vertices, firstVertexIndex, properties);					
				}
				++l;
			}
			
			// TODO ignore bone data by checking the name (MIKOTO bone 'bone:' or 'anchor')
			//ignore bone data
			//if (_meshData.material)
			//{
				_meshDataList.push(_meshData);
			//}
			//else
			//{
				//trace("mesh data material is null. ignoring bone data??");
			//}

			//TODO : do we need this?
			/*
			// Resolve parent-child relationship.
			var depth:int;
			try
			{
				depth = parseInt(properties["depth"]);
			}
			catch (e:Error)
			{
				depth = 0;
			}
			var parentMesh:ObjectContainer3D = _prevMesh;
			if (depth <= 0)
			{
				parentMesh = this;
				depth = 0;
			}
			else
			{
				while (depth <= _prevDepth)
				{
					parentMesh = ObjectContainer3D(parentMesh).parent;
					--_prevDepth;
				}
			}
			parentMesh.addChild(mesh);
			_prevMesh = mesh;
			_prevDepth = depth;
			*/

			return faceEndLine;
		}
		
		private function parseFace(line:String, vertices:Array, vertexOffset:int, properties:Dictionary):void
		{
			var vstr:String = getParam(line, "V");
			var mstr:String = getParam(line, "M");
			var uvstr:String = getParam(line, "UV");

			var v:Array = (vstr != null) ? vstr.match(/\d+/g) : [];
			var uv:Array = (uvstr != null) ? uvstr.match(/-?\d+\.\d+/g) : [];
			var a:Vertex;
			var b:Vertex;
			var c:Vertex;
			var d:Vertex;
			var materialData:MaterialData = null;
			var meshmaterialData:MeshMaterialData = null;
			var uvA:UV;
			var uvB:UV;
			var uvC:UV;
			var uvD:UV;
			var mirrorAxis:int;
			
			var symbolName:String 
			var temp:MeshMaterialData;
			var isFound:Boolean;
			var i:int;			
			if (v.length == 3)  // 3-vertices face
			{
				// get vertices
				c = vertices[parseInt(v[0]) + vertexOffset];
				b = vertices[parseInt(v[1]) + vertexOffset];
				a = vertices[parseInt(v[2]) + vertexOffset];

				// get material
				if (mstr != null)
				{
					materialData = _materialNames[parseInt(mstr)];					
					symbolName = materialData.name;
					meshmaterialData = getMeshMaterialData(symbolName, materialData);
				}

				// create UV
				if (uv.length != 0)
				{
					//uvC = new UV(parseFloat(uv[0]), parseFloat(uv[1]));
					//uvB = new UV(parseFloat(uv[2]), parseFloat(uv[3]));
					//uvA = new UV(parseFloat(uv[4]), parseFloat(uv[5]));					
					uvC = new UV(parseFloat(uv[0]), 1.0 - parseFloat(uv[1]));
					uvB = new UV(parseFloat(uv[2]), 1.0 - parseFloat(uv[3]));
					uvA = new UV(parseFloat(uv[4]), 1.0 - parseFloat(uv[5]));										
				}
				else  // no UV data
				{
					uvC = new UV(0, 1)
					uvB = new UV(1, 0)
					uvA = new UV(0, 0);
				}
				
				addFace(a, b, c, uvA, uvB, uvC, meshmaterialData);

				// handle mirror face
				if (properties["mirror"] == "1")
				{
					mirrorAxis = parseInt(properties["mirror_axis"]);
					
					// create mirror vertices
					a = mirrorVertex(a, mirrorAxis);
					b = mirrorVertex(b, mirrorAxis);
					c = mirrorVertex(c, mirrorAxis);
					vertices.push(a);
					vertices.push(b);
					vertices.push(c);
					
					addFace(c, b, a, uvC, uvB, uvA, meshmaterialData);
				}
			}
			else if (v.length == 4)  // 4-vertices face
			{
				// get vertices
				d = vertices[parseInt(v[0]) + vertexOffset];
				c = vertices[parseInt(v[1]) + vertexOffset];
				b = vertices[parseInt(v[2]) + vertexOffset];
				a = vertices[parseInt(v[3]) + vertexOffset];

				// get material
				if (mstr != null)
				{
					materialData = _materialNames[parseInt(mstr)];										
					symbolName = materialData.name;
					meshmaterialData = getMeshMaterialData(symbolName, materialData);
				}

				// create UV
				if (uv.length != 0)
				{
					//uvD = new UV(parseFloat(uv[0]), parseFloat(uv[1]));
					//uvC = new UV(parseFloat(uv[2]), parseFloat(uv[3]));
					//uvB = new UV(parseFloat(uv[4]), parseFloat(uv[5]));
					//uvA = new UV(parseFloat(uv[6]), parseFloat(uv[7]));
					uvD = new UV(parseFloat(uv[0]), 1.0 - parseFloat(uv[1]));
					uvC = new UV(parseFloat(uv[2]), 1.0 - parseFloat(uv[3]));
					uvB = new UV(parseFloat(uv[4]), 1.0 - parseFloat(uv[5]));
					uvA = new UV(parseFloat(uv[6]), 1.0 - parseFloat(uv[7]));					
				}
				else  // no UV
				{
					//uvD = new UV(1, 1);
					//uvC = new UV(0, 1);
					//uvB = new UV(1, 0);
					//uvA = new UV(0, 0);
					uvD = new UV(0, 0);
					uvC = new UV(1, 0);
					uvB = new UV(0, 1);
					uvA = new UV(1, 1);					
				}

				addFace(a, b, c, uvA, uvB, uvC, meshmaterialData);
				addFace(c, d, a, uvC, uvD, uvA, meshmaterialData);

				// handle mirror faces
				if (properties["mirror"] == "1")
				{
					mirrorAxis = parseInt(properties["mirror_axis"]);
					
					// create mirror vertices
					a = mirrorVertex(a, mirrorAxis);
					b = mirrorVertex(b, mirrorAxis);
					c = mirrorVertex(c, mirrorAxis);
					d = mirrorVertex(d, mirrorAxis);
					vertices.push(a);
					vertices.push(b);
					vertices.push(c);
					vertices.push(d);
					
					addFace(c, b, a, uvC, uvB, uvA, meshmaterialData);
					addFace(a, d, c, uvA, uvD, uvC, meshmaterialData);
				}
			}
		}

		private function getMeshMaterialData(symbolName:String, materialData:MaterialData):MeshMaterialData
		{
			var meshmaterialData:MeshMaterialData;
			var materialArray:Array = _meshData.geometry.materials;
			
			/*
			// check if the materialdata is already exist in symbol library
			if (_symbolLibrary[symbolName] != materialData)  // not exist before
			{
				Debug.trace("registering the symbol " + symbolName);
				
				// register symbol
				_symbolLibrary[symbolName] = materialData;						
				
				meshmaterialData = new MeshMaterialData();
				meshmaterialData.symbol = symbolName;										
				materialArray.push(meshmaterialData);
			}
			else  // has exist before
			{
				// get the meshmaterialdata form the materials array
				var isFound:Boolean = false;
				var i:int = 0;
				var numOfMaterials:int = materialArray.length;
				while ((!isFound) && (i < numOfMaterials))
				{
					meshmaterialData = materialArray[i] as MeshMaterialData;
					if (meshmaterialData.symbol == symbolName)
					{
						isFound = true;
					}
					i++;
				}
				
				// DEBUG
				if (!isFound)
				{
					Debug.trace("Error! MeshMaterialData is null for symbol '" + 
						symbolName +"'. This should not happen. numOfMaterial=" + 
						numOfMaterials + "._symbolLibrary[symbolName]=" + 
						_symbolLibrary[symbolName].name);
				}
			}			
			*/

			// get the meshmaterialdata form the materials array
			var isFound:Boolean = false;
			var i:int = 0;
			var numOfMaterials:int = materialArray.length;
			while ((!isFound) && (i < numOfMaterials))
			{
				meshmaterialData = materialArray[i] as MeshMaterialData;
				if (meshmaterialData.symbol == symbolName)
				{
					isFound = true;
				}
				i++;
			}
			
			// DEBUG
			if (!isFound)
			{
				_symbolLibrary[symbolName] = materialData;						
				
				meshmaterialData = new MeshMaterialData();
				meshmaterialData.symbol = symbolName;										
				materialArray.push(meshmaterialData);				
			}
			
			return meshmaterialData;			
		}
		
		private static function mirrorVertex(v:Vertex, axis:int):Vertex
		{
			return new Vertex(
				((axis & 1) != 0) ? -v.x : v.x, 
				((axis & 2) != 0) ? -v.y : v.y, 
				((axis & 4) != 0) ? -v.z : v.z);
		}

		private static function getChunkLine(lines:Array, chunkName:String, startLine:int = 0):int
		{
			for (var i:uint = startLine; i < lines.length; ++i)
			{
				if (lines[i].indexOf(chunkName) != -1)
				{
					return int(i);
				}
			}
			return -1;
		}
		
		private static function getParam(line:String, paramName:String):String
		{
			var prefix:String = paramName + "(";
			var prefixLen:int = prefix.length;

			var begin:int = line.indexOf(prefix, 0);

			if (begin == -1)
				return null;

			var end:int = line.indexOf(")", begin + prefixLen);

			if (end == -1)
				return null;

			return line.substring(begin + prefixLen, end);
		}
		
		private var n:int = -1;		

		private function addFace(v0:Vertex, v1:Vertex, v2:Vertex, uv0:UV, uv1:UV, uv2:UV, 
			meshMaterialData:MeshMaterialData):void
		{
			//_geometryData.vertices.push(-v0.x, -v0.y, v0.z, -v1.x, -v1.y, v1.z, -v2.x, -v2.y, v2.z);
			var vertices:Array = _geometryData.vertices;
			vertices.push(new Vertex(-v0.x, -v0.y, v0.z));
			vertices.push(new Vertex(-v1.x, -v1.y, v1.z));
			vertices.push(new Vertex(-v2.x, -v2.y, v2.z));
			
			//_geometryData.uvs.push(uv0.x, uv0.y, 1, uv1.x, uv1.y, 1, uv2.x, uv2.y, 1);
			var uvs:Array = _geometryData.uvs;
			uvs.push(new UV(uv0.u, uv0.v));
			uvs.push(new UV(uv1.u, uv1.v));
			uvs.push(new UV(uv2.u, uv2.v));
			
			n += 3;
			
			// create facedata
			//_geometryData.indices.push(n, n - 1, n - 2);
			//_geometryData.faceLengths.push(3);			
			var facedata:FaceData = new FaceData();
			facedata.uv0 = facedata.v0 = n - 2;
			facedata.uv1 = facedata.v1 = n - 1;
			facedata.uv2 = facedata.v2 = n;
			facedata.visible = true;			
			
			var faceIndex:int = _geometryData.faces.push(facedata);			
			faceIndex--;
			
			// handle meshMaterialData
			if (meshMaterialData != null)
			{				
				// insert facedata to this materialdata
				meshMaterialData.faceList.push(faceIndex);
			}			
			else
			{
				Debug.trace("MeshMaterialData is NULL during adding faces");
			}
		}

		
        /** @private */
        protected override function getFileType():String
        {
        	return "mqo";
        }
		
		public var charset:String = "shift_jis";

		/**
		 * Controls the use of shading materials when color textures are encountered. Defaults to false.
		 */
		public var shading:Boolean = false;

		/**
		 * A scaling factor for all geometry in the model. Defaults to 1.
		 */
		public var scaling:Number = 1;

		function Metasequoia (init:Object = null)
		{
			super(init);
			
			scaling = ini.getNumber("scaling", 1);
			
			_container = new ObjectContainer3D(ini);			
			_container.name = "metasequoia";
			
			_container.materialLibrary = _materialLibrary;
			_container.geometryLibrary = _geometryLibrary;
			
            //init = Init.parse(init);
			_filename = ini.getString("filename", "");
			
            //scaling = init.getNumber("scaling", 1)*100;
			//materials = ini.getObject("materials") || { };			
            //mesh = (container = new Mesh(ini)) as Mesh;
			
			binary = true;
		}
		
		/**
		 * Creates a 3d container object from the raw binary data of a 3ds file.
		 * 
		 * @param	data				The binary data of a loaded file.
		 * @param	init	[optional]	An initialisation object for specifying default instance properties.
		 * 
		 * @return						A 3d container object representation of the MQO file.
		 */
        public static function parse(data:*, init:Object = null):ObjectContainer3D
        {
        	return Loader3D.parse(data, Metasequoia, init).handle as ObjectContainer3D;
        }
		
    	/**
    	 * Loads and parses a MQO file into a 3d container object.
    	 * 
    	 * @param	url					The url location of the file to load.
    	 * @param	init	[optional]	An initialisation object for specifying default instance properties.
    	 * 
    	 * @return						A 3d loader object that can be used as a placeholder in a scene while the file is loading.
    	 */
        public static function load(url:String, init:Object = null):Loader3D
        {
            return Loader3D.load(url, Metasequoia, init);
        }
	}
}
